@@ -117,7 +117,7 @@ impl From<SigningCertificate> for CodeSignerValue {
 
 impl CodeSignerValue {
     fn signer(&self, label: &str) -> Result<std::sync::MutexGuard<Signer>, ValueError> {
-        self.inner.lock().map_err(|e| {
+        self.inner.try_lock().map_err(|e| {
             ValueError::Runtime(RuntimeError {
                 code: "TUGGER_CODE_SIGNING",
                 message: format!("error obtaining signer lock: {:?}", e),
@@ -574,10 +574,7 @@ pub fn handle_file_manifest_signable_events(
 
         let entry = if let Some(output) = response.output {
             if let SignedOutput::Memory(data) = output {
-                FileEntry {
-                    data: data.into(),
-                    executable: entry.executable,
-                }
+                FileEntry::new_from_data(data, entry.is_executable())
             } else {
                 return Err(anyhow!("SignedOutput::Memory should have been forced"));
             }
