@@ -9,6 +9,7 @@ use {
         file_manifest::FileManifestValue,
     },
     anyhow::{anyhow, Context},
+    apple_bundles::MacOsApplicationBundleBuilder,
     starlark::{
         environment::TypeValues,
         eval::call_stack::CallStack,
@@ -26,10 +27,12 @@ use {
         get_context_value, optional_str_arg, EnvironmentContext, ResolvedTarget,
         ResolvedTargetValue, RunMode,
     },
-    std::path::PathBuf,
-    tugger_apple_bundle::MacOsApplicationBundleBuilder,
+    std::{
+        convert::TryFrom,
+        path::{Path, PathBuf},
+    },
     tugger_code_signing::SigningDestination,
-    tugger_file_manifest::FileData,
+    tugger_file_manifest::FileEntry,
 };
 
 fn error_context<F, T>(label: &str, f: F) -> Result<T, ValueError>
@@ -45,7 +48,7 @@ where
     })
 }
 
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 pub struct MacOsApplicationBundleBuilderValue {
     pub inner: MacOsApplicationBundleBuilder,
 }
@@ -70,15 +73,20 @@ impl MacOsApplicationBundleBuilderValue {
 
     pub fn add_icon(&mut self, path: String) -> ValueResult {
         error_context("MacOsApplicationBundleBuilder.add_icon()", || {
-            self.inner.add_icon(FileData::from(PathBuf::from(path)))
+            self.inner
+                .add_icon(FileEntry::try_from(PathBuf::from(path))?)
         })?;
 
         Ok(Value::new(NoneType::None))
     }
 
     pub fn add_manifest(&mut self, manifest: FileManifestValue) -> ValueResult {
-        error_context("MacOsApplicationBundleBuilder.add_manifest()", || {
-            for (path, entry) in manifest.manifest.iter_entries() {
+        const LABEL: &str = "MacOsApplicationBundleBuilder.add_manifest()";
+
+        let manifest = manifest.inner(LABEL)?;
+
+        error_context(LABEL, || {
+            for (path, entry) in manifest.iter_entries() {
                 self.inner
                     .add_file(PathBuf::from("Contents").join(path), entry.clone())
                     .with_context(|| format!("adding {}", path.display()))?;
@@ -91,17 +99,21 @@ impl MacOsApplicationBundleBuilderValue {
     }
 
     pub fn add_macos_file(&mut self, content: FileContentValue, path: Value) -> ValueResult {
+        const LABEL: &str = "MacOsApplicationBundleBuilder.add_macos_file()";
+
         let path = optional_str_arg("path", &path)?;
 
+        let inner = content.inner(LABEL)?;
+
         error_context("MacOsApplicationBundleBuilder.add_macos_file()", || {
             let path = if let Some(path) = path {
                 PathBuf::from(path)
             } else {
-                PathBuf::from(&content.filename)
+                PathBuf::from(&inner.filename)
             };
 
             self.inner
-                .add_file_macos(path, content.content)
+                .add_file_macos(path, inner.content.clone())
                 .context("adding macOS file")
         })?;
 
@@ -109,8 +121,12 @@ impl MacOsApplicationBundleBuilderValue {
     }
 
     pub fn add_macos_manifest(&mut self, manifest: FileManifestValue) -> ValueResult {
-        error_context("MacOsApplicationBundleBuilder.add_macos_manifest()", || {
-            for (path, entry) in manifest.manifest.iter_entries() {
+        const LABEL: &str = "MacOsApplicationBundleBuilder.add_macos_manifest()";
+
+        let manifest = manifest.inner(LABEL)?;
+
+        error_context(LABEL, || {
+            for (path, entry) in manifest.iter_entries() {
                 self.inner
                     .add_file_macos(path, entry.clone())
                     .with_context(|| format!("adding {}", path.display()))?;
@@ -123,17 +139,21 @@ impl MacOsApplicationBundleBuilderValue {
     }
 
     pub fn add_resources_file(&mut self, content: FileContentValue, path: Value) -> ValueResult {
+        const LABEL: &str = "MacOsApplicationBundleBuilder.add_resources_file()";
+
         let path = optional_str_arg("path", &path)?;
 
-        error_context("MacOsApplicationBundleBuilder.add_resources_file()", || {
+        let inner = content.inner(LABEL)?;
+
+        error_context(LABEL, || {
             let path = if let Some(path) = path {
                 PathBuf::from(path)
             } else {
-                PathBuf::from(&content.filename)
+                PathBuf::from(&inner.filename)
             };
 
             self.inner
-                .add_file_resources(path, content.content)
+                .add_file_resources(path, inner.content.clone())
                 .context("adding resources file")
         })?;
 
@@ -141,18 +161,19 @@ impl MacOsApplicationBundleBuilderValue {
     }
 
     pub fn add_resources_manifest(&mut self, manifest: FileManifestValue) -> ValueResult {
-        error_context(
-            "MacOsApplicationBundleBuilder.add_resources_manifest()",
-            || {
-                for (path, entry) in manifest.manifest.iter_entries() {
-                    self.inner
-                        .add_file_resources(path, entry.clone())
-                        .with_context(|| format!("adding {}", path.display()))?;
-                }
+        const LABEL: &str = "MacOsApplicationBundleBuilder.add_resources_manifest()";
 
-                Ok(())
-            },
-        )?;
+        let manifest = manifest.inner(LABEL)?;
+
+        error_context(LABEL, || {
+            for (path, entry) in manifest.iter_entries() {
+                self.inner
+                    .add_file_resources(path, entry.clone())
+                    .with_context(|| format!("adding {}", path.display()))?;
+            }
+
+            Ok(())
+        })?;
 
         Ok(Value::new(NoneType::None))
     }
@@ -206,6 +227,42 @@ impl MacOsApplicationBundleBuilderValue {
         Ok(Value::new(NoneType::None))
     }
 
+    fn materialize_bundle(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        label: &'static str,
+        dest_dir: &Path,
+    ) -> Result<PathBuf, ValueError> {
+        let (bundle_path, filename) = error_context(label, || {
+            let bundle_path = self
+                .inner
+                .materialize_bundle(dest_dir)
+                .context("materializing bundle")?;
+
+            let filename = bundle_path
+                .file_name()
+                .ok_or_else(|| anyhow!("unable to resolve bundle file name"))?
+                .to_os_string();
+
+            Ok((bundle_path, filename))
+        })?;
+
+        let candidate = bundle_path.as_path().into();
+        let mut context = SigningContext::new(
+            label,
+            SigningAction::MacOsApplicationBunderCreation,
+            filename,
+            &candidate,
+        );
+        context.set_path(&bundle_path);
+        context.set_signing_destination(SigningDestination::Directory(bundle_path.clone()));
+
+        handle_signable_event(type_values, call_stack, context)?;
+
+        Ok(bundle_path)
+    }
+
     pub fn build(
         &self,
         type_values: &TypeValues,
@@ -221,31 +278,7 @@ impl MacOsApplicationBundleBuilderValue {
 
         let output_path = context.target_build_path(&target);
 
-        let (bundle_path, filename) = error_context(LABEL, || {
-            let bundle_path = self
-                .inner
-                .materialize_bundle(&output_path)
-                .context("materializing bundle")?;
-
-            let filename = bundle_path
-                .file_name()
-                .ok_or_else(|| anyhow!("unable to resolve bundle file name"))?
-                .to_os_string();
-
-            Ok((bundle_path, filename))
-        })?;
-
-        let candidate = bundle_path.as_path().into();
-        let mut context = SigningContext::new(
-            LABEL,
-            SigningAction::MacOsApplicationBunderCreation,
-            filename,
-            &candidate,
-        );
-        context.set_path(&bundle_path);
-        context.set_signing_destination(SigningDestination::Directory(bundle_path.clone()));
-
-        handle_signable_event(type_values, call_stack, context)?;
+        let bundle_path = self.materialize_bundle(type_values, call_stack, LABEL, &output_path)?;
 
         Ok(Value::new(ResolvedTargetValue {
             inner: ResolvedTarget {
@@ -254,6 +287,26 @@ impl MacOsApplicationBundleBuilderValue {
             },
         }))
     }
+
+    pub fn write_to_directory(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        path: String,
+    ) -> ValueResult {
+        const LABEL: &str = "MacOsApplicationBundleBuilder.write_to_directory()";
+
+        let context_value = get_context_value(type_values)?;
+        let context = context_value
+            .downcast_ref::<EnvironmentContext>()
+            .ok_or(ValueError::IncorrectParameterType)?;
+
+        let dest_dir = context.resolve_path(path);
+
+        let bundle_path = self.materialize_bundle(type_values, call_stack, LABEL, &dest_dir)?;
+
+        Ok(Value::from(format!("{}", bundle_path.display())))
+    }
 }
 
 starlark_module! { macos_application_bundle_builder_module =>
@@ -262,19 +315,16 @@ starlark_module! { macos_application_bundle_builder_module =>
         MacOsApplicationBundleBuilderValue::new_from_args(bundle_name)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_icon(this, path: String) {
         let mut this = this.downcast_mut::<MacOsApplicationBundleBuilderValue>().unwrap().unwrap();
         this.add_icon(path)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_manifest(this, manifest: FileManifestValue) {
         let mut this = this.downcast_mut::<MacOsApplicationBundleBuilderValue>().unwrap().unwrap();
         this.add_manifest(manifest)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_macos_file(
         this,
         content: FileContentValue,
@@ -285,13 +335,11 @@ starlark_module! { macos_application_bundle_builder_module =>
         this.add_macos_file(content, path)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_macos_manifest(this, manifest: FileManifestValue) {
         let mut this = this.downcast_mut::<MacOsApplicationBundleBuilderValue>().unwrap().unwrap();
         this.add_macos_manifest(manifest)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_resources_file(
         this,
         content: FileContentValue,
@@ -301,19 +349,16 @@ starlark_module! { macos_application_bundle_builder_module =>
         this.add_resources_file(content, path)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.add_resources_manifest(this, manifest: FileManifestValue) {
         let mut this = this.downcast_mut::<MacOsApplicationBundleBuilderValue>().unwrap().unwrap();
         this.add_resources_manifest(manifest)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.set_info_plist_key(this, key: String, value: Value) {
         let mut this = this.downcast_mut::<MacOsApplicationBundleBuilderValue>().unwrap().unwrap();
         this.set_info_plist_key(key, value)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.set_info_plist_required_keys(
         this,
         display_name: String,
@@ -326,16 +371,20 @@ starlark_module! { macos_application_bundle_builder_module =>
         this.set_info_plist_required_keys(display_name, identifier, version, signature, executable)
     }
 
-    #[allow(non_snake_case)]
     MacOsApplicationBundleBuilder.build(env env, call_stack cs, this, target: String) {
         let this = this.downcast_ref::<MacOsApplicationBundleBuilderValue>().unwrap();
         this.build(env, cs, target)
     }
+
+    MacOsApplicationBundleBuilder.write_to_directory(env env, call_stack cs, this, path: String) {
+        let this = this.downcast_ref::<MacOsApplicationBundleBuilderValue>().unwrap();
+        this.write_to_directory(env, cs, path)
+    }
 }
 
 #[cfg(test)]
 mod tests {
-    use {super::*, crate::starlark::testutil::*, anyhow::Result};
+    use {super::*, crate::starlark::testutil::*, anyhow::Result, tugger_common::testutil::*};
 
     #[test]
     fn constructor() -> Result<()> {
@@ -420,4 +469,27 @@ mod tests {
 
         Ok(())
     }
+
+    #[test]
+    fn write_to_directory() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        env.eval("builder = MacOsApplicationBundleBuilder('myapp')")?;
+        env.eval(
+            "builder.add_resources_file(FileContent(filename = 'file', content = 'content'))",
+        )?;
+
+        let dest_dir = DEFAULT_TEMP_DIR
+            .path()
+            .join("macos-application-bundle-builder-write-to-directory");
+        let dest_dir_s = dest_dir.to_string_lossy().replace('\\', "/");
+
+        let path_value = env.eval(&format!("builder.write_to_directory('{}')", dest_dir_s))?;
+        assert_eq!(path_value.get_type(), "string");
+
+        let path = PathBuf::from(path_value.to_string());
+        assert!(path.is_dir());
+
+        Ok(())
+    }
 }
