@@ -8,10 +8,11 @@ use {
             handle_file_manifest_signable_events, handle_signable_event, SigningAction,
             SigningContext,
         },
+        file_content::FileContentWrapper,
         file_manifest::FileManifestValue,
         wix_msi_builder::WiXMsiBuilderValue,
     },
-    anyhow::{Context, Result},
+    anyhow::{anyhow, Context, Result},
     starlark::{
         environment::TypeValues,
         eval::call_stack::CallStack,
@@ -29,7 +30,10 @@ use {
         get_context_value, optional_dict_arg, optional_str_arg, EnvironmentContext, ResolvedTarget,
         ResolvedTargetValue, RunMode,
     },
-    std::{convert::TryFrom, path::Path},
+    std::{
+        convert::TryFrom,
+        path::{Path, PathBuf},
+    },
     tugger_code_signing::SigningDestination,
     tugger_file_manifest::{FileEntry, FileManifest},
     tugger_wix::{WiXInstallerBuilder, WiXSimpleMsiBuilder, WxsBuilder},
@@ -63,6 +67,7 @@ impl TypedValue for WiXInstallerValue {
 
     fn get_attr(&self, attribute: &str) -> ValueResult {
         Ok(match attribute {
+            "arch" => Value::from(self.inner.arch()),
             "install_files_root_directory_id" => {
                 Value::from(self.inner.install_files_root_directory_id())
             }
@@ -82,12 +87,15 @@ impl TypedValue for WiXInstallerValue {
     fn has_attr(&self, attribute: &str) -> Result<bool, ValueError> {
         Ok(matches!(
             attribute,
-            "install_files_root_directory_id" | "install_files_wxs_path"
+            "arch" | "install_files_root_directory_id" | "install_files_wxs_path"
         ))
     }
 
     fn set_attr(&mut self, attribute: &str, value: Value) -> Result<(), ValueError> {
         match attribute {
+            "arch" => {
+                self.inner.set_arch(value.to_string());
+            }
             "install_files_root_directory_id" => {
                 self.inner
                     .set_install_files_root_directory_id(value.to_string());
@@ -109,14 +117,18 @@ impl TypedValue for WiXInstallerValue {
 }
 
 impl WiXInstallerValue {
-    fn new_from_args(type_values: &TypeValues, id: String, filename: String) -> ValueResult {
+    fn new_from_args(
+        type_values: &TypeValues,
+        id: String,
+        filename: String,
+        arch: String,
+    ) -> ValueResult {
         let build_context_value = get_context_value(type_values)?;
         let context = build_context_value
             .downcast_ref::<EnvironmentContext>()
             .ok_or(ValueError::IncorrectParameterType)?;
 
-        // TODO grab target triple properly.
-        let builder = WiXInstallerBuilder::new(id, env!("HOST").to_string(), context.build_path());
+        let builder = WiXInstallerBuilder::new(id, arch, context.build_path());
 
         Ok(Value::new(WiXInstallerValue {
             inner: builder,
@@ -125,9 +137,13 @@ impl WiXInstallerValue {
     }
 
     fn add_build_files(&mut self, manifest: FileManifestValue) -> ValueResult {
-        error_context("WiXInstaller.add_build_files()", || {
+        const LABEL: &str = "WiXInstaller.add_build_files()";
+
+        let manifest = manifest.inner(LABEL)?;
+
+        error_context(LABEL, || {
             self.inner
-                .add_extra_build_files(&manifest.manifest)
+                .add_extra_build_files(&manifest)
                 .context("adding extra build files from FileManifest")
         })?;
 
@@ -196,11 +212,15 @@ impl WiXInstallerValue {
         call_stack: &mut CallStack,
         manifest: FileManifestValue,
     ) -> ValueResult {
+        const LABEL: &str = "WixInstaller.add_install_files()";
+
+        let manifest = manifest.inner(LABEL)?;
+
         self.add_install_files_from_manifest(
             type_values,
             call_stack,
             "WiXInstaller.add_install_files()",
-            &manifest.manifest,
+            &manifest,
         )
     }
 
@@ -231,9 +251,13 @@ impl WiXInstallerValue {
     }
 
     fn add_msi_builder(&mut self, builder: WiXMsiBuilderValue) -> ValueResult {
-        error_context("WiXInstaller.add_msi_builder()", || {
-            builder
-                .inner
+        const LABEL: &str = "WiXInstaller.add_msi_builder()";
+
+        let inner = builder.inner(LABEL)?;
+
+        error_context(LABEL, || {
+            inner
+                .builder
                 .add_to_installer_builder(&mut self.inner)
                 .context("adding WiXInstallerBuilder")
         })?;
@@ -249,7 +273,11 @@ impl WiXInstallerValue {
         product_manufacturer: String,
         program_files: FileManifestValue,
     ) -> ValueResult {
-        error_context("WiXInstaller.add_simple_installer()", || {
+        const LABEL: &str = "WiXInstaller.add_simple_installer()";
+
+        let manifest = program_files.inner(LABEL)?;
+
+        error_context(LABEL, || {
             let mut builder = WiXSimpleMsiBuilder::new(
                 &id_prefix,
                 &product_name,
@@ -257,7 +285,7 @@ impl WiXInstallerValue {
                 &product_manufacturer,
             );
             builder
-                .add_program_files_manifest(&program_files.manifest)
+                .add_program_files_manifest(&manifest)
                 .context("adding program files manifest")?;
 
             builder
@@ -298,33 +326,42 @@ impl WiXInstallerValue {
         Ok(Value::new(NoneType::None))
     }
 
-    fn build(
+    fn materialize(
         &mut self,
         type_values: &TypeValues,
         call_stack: &mut CallStack,
-        target: String,
-    ) -> ValueResult {
-        let context_value = get_context_value(type_values)?;
-        let context = context_value
-            .downcast_ref::<EnvironmentContext>()
-            .ok_or(ValueError::IncorrectParameterType)?;
+        label: &'static str,
+        dest_dir: &Path,
+    ) -> Result<PathBuf, ValueError> {
+        let logger = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
 
-        let output_path = context.target_build_path(&target);
-        let installer_path = output_path.join(&self.filename);
+            context.logger().clone()
+        };
+
+        let installer_path = dest_dir.join(&self.filename);
+
+        error_context(label, || {
+            // Update the build directory to use the destination directory.
+            // This helps prevent conflicts between parallel builds, as the default
+            // build path is hard-coded.
+            self.inner.set_build_path(dest_dir);
 
-        error_context("WiXInstaller.build()", || {
             self.inner
                 .add_files_manifest_wxs()
                 .context("generating install_files manifest wxs")?;
 
             self.inner
-                .build(context.logger(), &installer_path)
+                .build(&logger, &installer_path)
                 .context("building")
         })?;
 
         let candidate = installer_path.as_path().into();
         let mut context = SigningContext::new(
-            "build()",
+            label,
             SigningAction::WindowsInstallerCreation,
             &self.filename,
             &candidate,
@@ -334,6 +371,67 @@ impl WiXInstallerValue {
 
         handle_signable_event(type_values, call_stack, context)?;
 
+        Ok(installer_path)
+    }
+
+    fn materialize_temp_dir(
+        &mut self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        label: &'static str,
+    ) -> Result<(FileEntry, String), ValueError> {
+        let build_path = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
+
+            context.build_path().to_path_buf()
+        };
+
+        let dest_dir = error_context(label, || {
+            tempfile::Builder::new()
+                .prefix("wix-installer-")
+                .tempdir_in(&build_path)
+                .context("creating temp directory")
+        })?;
+
+        let installer_path = self.materialize(type_values, call_stack, label, dest_dir.path())?;
+
+        let entry = FileEntry::new_from_path(&installer_path, false);
+
+        let (entry, filename) = error_context(label, || {
+            let entry = entry
+                .to_memory()
+                .context("converting FileEntry to in-memory")?;
+
+            let filename = installer_path
+                .file_name()
+                .ok_or_else(|| anyhow!("unable to resolve file name of generated installer"))?;
+
+            Ok((entry, filename.to_string_lossy().to_string()))
+        })?;
+
+        Ok((entry, filename))
+    }
+
+    fn build(
+        &mut self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        target: String,
+    ) -> ValueResult {
+        const LABEL: &str = "WixInstaller.build()";
+
+        let context_value = get_context_value(type_values)?;
+        let context = context_value
+            .downcast_ref::<EnvironmentContext>()
+            .ok_or(ValueError::IncorrectParameterType)?;
+
+        let output_path = context.target_build_path(&target);
+
+        let installer_path = self.materialize(type_values, call_stack, LABEL, &output_path)?;
+
         Ok(Value::new(ResolvedTargetValue {
             inner: ResolvedTarget {
                 run_mode: RunMode::Path {
@@ -350,12 +448,59 @@ impl WiXInstallerValue {
 
         Ok(Value::new(NoneType::None))
     }
+
+    #[allow(clippy::wrong_self_convention)]
+    fn to_file_content(
+        &mut self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+    ) -> ValueResult {
+        const LABEL: &str = "WiXInstaller.to_file_content()";
+
+        let (entry, filename) = self.materialize_temp_dir(type_values, call_stack, LABEL)?;
+
+        Ok(FileContentWrapper {
+            content: entry,
+            filename,
+        }
+        .into())
+    }
+
+    fn write_to_directory(
+        &mut self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        path: String,
+    ) -> ValueResult {
+        const LABEL: &str = "WiXInstaller.write_to_directory()";
+
+        let dest_dir = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
+
+            context.resolve_path(path)
+        };
+
+        let (entry, filename) = self.materialize_temp_dir(type_values, call_stack, LABEL)?;
+
+        let installer_path = dest_dir.join(&filename);
+
+        error_context(LABEL, || {
+            entry
+                .write_to_path(&installer_path)
+                .with_context(|| format!("writing installer to {}", installer_path.display()))
+        })?;
+
+        Ok(Value::from(format!("{}", installer_path.display())))
+    }
 }
 
 starlark_module! { wix_installer_module =>
     #[allow(non_snake_case)]
-    WiXInstaller(env env, id: String, filename: String) {
-        WiXInstallerValue::new_from_args(env, id, filename)
+    WiXInstaller(env env, id: String, filename: String, arch: String = "x64".to_string()) {
+        WiXInstallerValue::new_from_args(env, id, filename, arch)
     }
 
     WiXInstaller.add_build_file(
@@ -428,11 +573,23 @@ starlark_module! { wix_installer_module =>
         let mut this = this.downcast_mut::<WiXInstallerValue>().unwrap().unwrap();
         this.set_variable(key, value)
     }
+
+    WiXInstaller.to_file_content(env env, call_stack cs, this) {
+        let mut this = this.downcast_mut::<WiXInstallerValue>().unwrap().unwrap();
+        this.to_file_content(env, cs)
+    }
+
+    WiXInstaller.write_to_directory(env env, call_stack cs, this, path: String) {
+        let mut this = this.downcast_mut::<WiXInstallerValue>().unwrap().unwrap();
+        this.write_to_directory(env, cs, path)
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use {super::*, crate::starlark::testutil::*, anyhow::Result};
+    #[cfg(windows)]
+    use {crate::starlark::file_content::FileContentValue, tugger_common::testutil::*};
 
     #[test]
     fn test_constructor() -> Result<()> {
@@ -441,6 +598,24 @@ mod tests {
         let installer = env.eval("WiXInstaller('myapp', 'ignored')")?;
         assert_eq!(installer.get_type(), WiXInstallerValue::TYPE);
 
+        let installer_value = env.eval("WiXInstaller('myapp', 'ignored', arch='arch')")?;
+        let installer = installer_value.downcast_ref::<WiXInstallerValue>().unwrap();
+        assert_eq!(installer.inner.arch(), "arch");
+
+        Ok(())
+    }
+
+    #[test]
+    fn test_attributes() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        env.eval("i = WiXInstaller('myapp', 'ignored')")?;
+        let arch = env.eval("i.arch")?;
+        assert_eq!(arch.to_string(), "x64");
+        env.eval("i.arch = 'x86'")?;
+        let arch = env.eval("i.arch")?;
+        assert_eq!(arch.to_string(), "x86");
+
         Ok(())
     }
 
@@ -534,4 +709,42 @@ mod tests {
 
         Ok(())
     }
+
+    #[cfg(windows)]
+    #[test]
+    fn to_file_content() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        env.eval("installer = WiXInstaller('myapp', 'myapp.msi')")?;
+        env.eval(
+            "installer.add_simple_installer('myapp', 'myapp', '0.1', 'author', FileManifest())",
+        )?;
+        let value = env.eval("installer.to_file_content()")?;
+        assert_eq!(value.get_type(), FileContentValue::TYPE);
+
+        Ok(())
+    }
+
+    #[cfg(windows)]
+    #[test]
+    fn write_to_directory() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        let dest_dir = DEFAULT_TEMP_DIR
+            .path()
+            .join("wix-installer-write-to-directory");
+        let dest_dir_s = dest_dir.to_string_lossy().replace('\\', "/");
+
+        env.eval("installer = WiXInstaller('myapp', 'myapp.msi')")?;
+        env.eval(
+            "installer.add_simple_installer('myapp', 'myapp', '0.1', 'author', FileManifest())",
+        )?;
+        let value = env.eval(&format!("installer.write_to_directory('{}')", dest_dir_s))?;
+        assert_eq!(value.get_type(), "string");
+
+        let path = PathBuf::from(value.to_string());
+        assert!(path.exists());
+
+        Ok(())
+    }
 }
