@@ -8,9 +8,10 @@ use {
             handle_file_manifest_signable_events, handle_signable_event, SigningAction,
             SigningContext,
         },
+        file_content::FileContentWrapper,
         file_manifest::FileManifestValue,
     },
-    anyhow::{Context, Result},
+    anyhow::{anyhow, Context, Result},
     starlark::{
         environment::TypeValues,
         eval::call_stack::CallStack,
@@ -27,8 +28,13 @@ use {
     starlark_dialect_build_targets::{
         get_context_value, EnvironmentContext, ResolvedTarget, ResolvedTargetValue, RunMode,
     },
-    std::convert::TryFrom,
+    std::{
+        convert::TryFrom,
+        path::{Path, PathBuf},
+        sync::{Arc, Mutex, MutexGuard},
+    },
     tugger_code_signing::SigningDestination,
+    tugger_file_manifest::FileEntry,
     tugger_windows::VcRedistributablePlatform,
     tugger_wix::WiXSimpleMsiBuilder,
 };
@@ -46,13 +52,17 @@ where
     })
 }
 
+pub struct WiXMsiBuilderWrapper {
+    pub builder: WiXSimpleMsiBuilder,
+    /// Explicit filename to use for the built MSI.
+    pub msi_filename: Option<String>,
+    /// The target architecture we are building for.
+    pub arch: String,
+}
+
 #[derive(Clone)]
 pub struct WiXMsiBuilderValue {
-    pub inner: WiXSimpleMsiBuilder,
-    /// Explicit filename to use for the built MSI.
-    msi_filename: Option<String>,
-    /// The target architecture we are building for.
-    pub target_triple: String,
+    inner: Arc<Mutex<WiXMsiBuilderWrapper>>,
 }
 
 impl TypedValue for WiXMsiBuilderValue {
@@ -64,39 +74,41 @@ impl TypedValue for WiXMsiBuilderValue {
     }
 
     fn set_attr(&mut self, attribute: &str, value: Value) -> Result<(), ValueError> {
+        let mut inner = self.inner(&format!("{}.{}", Self::TYPE, &attribute))?;
+
         match attribute {
+            "arch" => {
+                inner.arch = value.to_string();
+            }
             "banner_bmp_path" => {
-                self.inner = self.inner.clone().banner_bmp_path(value.to_string());
+                inner.builder = inner.builder.clone().banner_bmp_path(value.to_string());
             }
             "dialog_bmp_path" => {
-                self.inner = self.inner.clone().dialog_bmp_path(value.to_string());
+                inner.builder = inner.builder.clone().dialog_bmp_path(value.to_string());
             }
             "eula_rtf_path" => {
-                self.inner = self.inner.clone().eula_rtf_path(value.to_string());
+                inner.builder = inner.builder.clone().eula_rtf_path(value.to_string());
             }
             "help_url" => {
-                self.inner = self.inner.clone().help_url(value.to_string());
+                inner.builder = inner.builder.clone().help_url(value.to_string());
             }
             "license_path" => {
-                self.inner = self.inner.clone().license_path(value.to_string());
+                inner.builder = inner.builder.clone().license_path(value.to_string());
             }
             "msi_filename" => {
-                self.msi_filename = Some(value.to_string());
+                inner.msi_filename = Some(value.to_string());
             }
             "package_description" => {
-                self.inner = self.inner.clone().package_description(value.to_string());
+                inner.builder = inner.builder.clone().package_description(value.to_string());
             }
             "package_keywords" => {
-                self.inner = self.inner.clone().package_keywords(value.to_string());
+                inner.builder = inner.builder.clone().package_keywords(value.to_string());
             }
             "product_icon_path" => {
-                self.inner = self.inner.clone().product_icon_path(value.to_string());
-            }
-            "target_triple" => {
-                self.target_triple = value.to_string();
+                inner.builder = inner.builder.clone().product_icon_path(value.to_string());
             }
             "upgrade_code" => {
-                self.inner = self.inner.clone().upgrade_code(value.to_string());
+                inner.builder = inner.builder.clone().upgrade_code(value.to_string());
             }
             attr => {
                 return Err(ValueError::OperationNotSupported {
@@ -117,8 +129,9 @@ impl WiXMsiBuilderValue {
         product_name: String,
         product_version: String,
         product_manufacturer: String,
+        arch: String,
     ) -> ValueResult {
-        let inner = WiXSimpleMsiBuilder::new(
+        let builder = WiXSimpleMsiBuilder::new(
             &id_prefix,
             &product_name,
             &product_version,
@@ -126,12 +139,24 @@ impl WiXMsiBuilderValue {
         );
 
         Ok(Value::new(WiXMsiBuilderValue {
-            inner,
-            msi_filename: None,
-            target_triple: env!("HOST").to_string(),
+            inner: Arc::new(Mutex::new(WiXMsiBuilderWrapper {
+                builder,
+                msi_filename: None,
+                arch,
+            })),
         }))
     }
 
+    pub fn inner(&self, label: &str) -> Result<MutexGuard<WiXMsiBuilderWrapper>, ValueError> {
+        self.inner.try_lock().map_err(|e| {
+            ValueError::Runtime(RuntimeError {
+                code: "TUGGER_WIX_MSI_BUILDER",
+                message: format!("error obtaining lock: {}", e),
+                label: label.to_string(),
+            })
+        })
+    }
+
     pub fn add_program_files_manifest(
         &mut self,
         type_values: &TypeValues,
@@ -140,16 +165,20 @@ impl WiXMsiBuilderValue {
     ) -> ValueResult {
         const LABEL: &str = "WiXMSIBuilder.add_program_files_manifest()";
 
+        let mut inner = self.inner(LABEL)?;
+        let manifest = manifest.inner(LABEL)?;
+
         error_context(LABEL, || {
             let manifest = handle_file_manifest_signable_events(
                 type_values,
                 call_stack,
-                &manifest.manifest,
+                &manifest,
                 LABEL,
                 SigningAction::WindowsInstallerFileAdded,
             )?;
 
-            self.inner
+            inner
+                .builder
                 .add_program_files_manifest(&manifest)
                 .context("adding program files manifest")
         })?;
@@ -162,11 +191,16 @@ impl WiXMsiBuilderValue {
         redist_version: String,
         platform: String,
     ) -> ValueResult {
-        error_context("WiXMSIBuilder.add_visual_cpp_redistributable()", || {
+        const LABEL: &str = "WiXMSIBuilder.add_visual_cpp_redistributable()";
+
+        let mut inner = self.inner(LABEL)?;
+
+        error_context(LABEL, || {
             let platform = VcRedistributablePlatform::try_from(platform.as_str())
                 .context("obtaining VcRedistributablePlatform from str")?;
 
-            self.inner
+            inner
+                .builder
                 .add_visual_cpp_redistributable(&redist_version, platform)
                 .context("adding Visual C++ redistributable")
         })?;
@@ -174,6 +208,96 @@ impl WiXMsiBuilderValue {
         Ok(Value::new(NoneType::None))
     }
 
+    pub fn materialize(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        label: &'static str,
+        build_dir: &Path,
+    ) -> Result<PathBuf, ValueError> {
+        let logger = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
+
+            context.logger().clone()
+        };
+
+        let msi_filename = self.msi_filename(label)?;
+        let inner = self.inner(label)?;
+
+        let msi_path = error_context(label, || {
+            let builder = inner
+                .builder
+                .to_installer_builder(&inner.arch, build_dir)
+                .context("converting WiXSimpleMSiBuilder to WiXInstallerBuilder")?;
+
+            let msi_path = build_dir.join(&msi_filename);
+
+            builder
+                .build(&logger, &msi_path)
+                .context("building WiXInstallerBuilder")?;
+
+            Ok(msi_path)
+        })?;
+
+        let candidate = msi_path.as_path().into();
+        let mut context = SigningContext::new(
+            label,
+            SigningAction::WindowsInstallerCreation,
+            &msi_filename,
+            &candidate,
+        );
+        context.set_path(&msi_path);
+        context.set_signing_destination(SigningDestination::File(msi_path.clone()));
+
+        handle_signable_event(type_values, call_stack, context)?;
+
+        Ok(msi_path)
+    }
+
+    fn materialize_temp_dir(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        label: &'static str,
+    ) -> Result<(FileEntry, String), ValueError> {
+        let build_path = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
+
+            context.build_path().to_path_buf()
+        };
+
+        let dest_dir = error_context(label, || {
+            tempfile::Builder::new()
+                .prefix("wix-msi-builder-")
+                .tempdir_in(&build_path)
+                .context("creating temp directory")
+        })?;
+
+        let installer_path = self.materialize(type_values, call_stack, label, dest_dir.path())?;
+
+        let entry = FileEntry::new_from_path(&installer_path, false);
+
+        let (entry, filename) = error_context(label, || {
+            let entry = entry
+                .to_memory()
+                .context("converting FileEntry to in-memory")?;
+
+            let filename = installer_path
+                .file_name()
+                .ok_or_else(|| anyhow!("unable to resolve file name of generated installer"))?;
+
+            Ok((entry, filename.to_string_lossy().to_string()))
+        })?;
+
+        Ok((entry, filename))
+    }
+
     pub fn build(
         &self,
         type_values: &TypeValues,
@@ -182,54 +306,79 @@ impl WiXMsiBuilderValue {
     ) -> ValueResult {
         const LABEL: &str = "WiXMSIBuilder.build()";
 
-        let context_value = get_context_value(type_values)?;
-        let context = context_value
-            .downcast_ref::<EnvironmentContext>()
-            .ok_or(ValueError::IncorrectParameterType)?;
+        let dest_dir = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
 
-        let (msi_path, output_path) = error_context(LABEL, || {
-            let output_path = context.target_build_path(&target);
+            context.target_build_path(&target)
+        };
 
-            let builder = self
-                .inner
-                .to_installer_builder(&self.target_triple, &output_path)
-                .context("converting WiXSimpleMSiBuilder to WiXInstallerBuilder")?;
-
-            let msi_path = output_path.join(self.msi_filename());
-
-            builder
-                .build(context.logger(), &msi_path)
-                .context("building WiXInstallerBuilder")?;
-
-            Ok((msi_path, output_path))
-        })?;
-
-        let candidate = msi_path.as_path().into();
-        let mut context = SigningContext::new(
-            "build()",
-            SigningAction::WindowsInstallerCreation,
-            self.msi_filename(),
-            &candidate,
-        );
-        context.set_path(&msi_path);
-        context.set_signing_destination(SigningDestination::File(msi_path.clone()));
-
-        handle_signable_event(type_values, call_stack, context)?;
+        let msi_path = self.materialize(type_values, call_stack, LABEL, &dest_dir)?;
 
         Ok(Value::new(ResolvedTargetValue {
             inner: ResolvedTarget {
                 run_mode: RunMode::Path { path: msi_path },
-                output_path,
+                output_path: dest_dir,
             },
         }))
     }
 
-    pub fn msi_filename(&self) -> String {
-        if let Some(filename) = &self.msi_filename {
+    pub fn msi_filename(&self, label: &str) -> Result<String, ValueError> {
+        let inner = self.inner(label)?;
+
+        Ok(if let Some(filename) = &inner.msi_filename {
             filename.clone()
         } else {
-            self.inner.default_msi_filename()
+            inner.builder.default_msi_filename()
+        })
+    }
+
+    pub fn to_file_content(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+    ) -> ValueResult {
+        const LABEL: &str = "WiXMSIBuilder.to_file_content()";
+
+        let (entry, filename) = self.materialize_temp_dir(type_values, call_stack, LABEL)?;
+
+        Ok(FileContentWrapper {
+            content: entry,
+            filename,
         }
+        .into())
+    }
+
+    fn write_to_directory(
+        &self,
+        type_values: &TypeValues,
+        call_stack: &mut CallStack,
+        path: String,
+    ) -> ValueResult {
+        const LABEL: &str = "WiXMSIBuilder.write_to_directory()";
+
+        let dest_dir = {
+            let context_value = get_context_value(type_values)?;
+            let context = context_value
+                .downcast_ref::<EnvironmentContext>()
+                .ok_or(ValueError::IncorrectParameterType)?;
+
+            context.resolve_path(path)
+        };
+
+        let (entry, filename) = self.materialize_temp_dir(type_values, call_stack, LABEL)?;
+
+        let installer_path = dest_dir.join(&filename);
+
+        error_context(LABEL, || {
+            entry
+                .write_to_path(&installer_path)
+                .with_context(|| format!("writing installer to {}", installer_path.display()))
+        })?;
+
+        Ok(Value::from(format!("{}", installer_path.display())))
     }
 }
 
@@ -239,18 +388,17 @@ starlark_module! { wix_msi_builder_module =>
         id_prefix: String,
         product_name: String,
         product_version: String,
-        product_manufacturer: String
+        product_manufacturer: String,
+        arch: String = "x64".to_string()
     ) {
-        WiXMsiBuilderValue::new_from_args(id_prefix, product_name, product_version, product_manufacturer)
+        WiXMsiBuilderValue::new_from_args(id_prefix, product_name, product_version, product_manufacturer, arch)
     }
 
-    #[allow(non_snake_case)]
     WiXMSIBuilder.add_program_files_manifest(env env, call_stack cs, this, manifest: FileManifestValue) {
         let mut this = this.downcast_mut::<WiXMsiBuilderValue>().unwrap().unwrap();
         this.add_program_files_manifest(env, cs, manifest)
     }
 
-    #[allow(non_snake_case)]
     WiXMSIBuilder.add_visual_cpp_redistributable(
         this,
         redist_version: String,
@@ -260,23 +408,36 @@ starlark_module! { wix_msi_builder_module =>
         this.add_visual_cpp_redistributable(redist_version, platform)
     }
 
-    #[allow(non_snake_case)]
     WiXMSIBuilder.build(env env, call_stack cs, this, target: String) {
         let this = this.downcast_ref::<WiXMsiBuilderValue>().unwrap();
         this.build(env, cs, target)
     }
+
+    WiXMSIBuilder.to_file_content(env env, call_stack cs, this) {
+        let this = this.downcast_ref::<WiXMsiBuilderValue>().unwrap();
+        this.to_file_content(env, cs)
+    }
+
+    WiXMSIBuilder.write_to_directory(env env, call_stack cs, this, path: String) {
+        let this = this.downcast_ref::<WiXMsiBuilderValue>().unwrap();
+        this.write_to_directory(env, cs, path)
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use {super::*, crate::starlark::testutil::*};
+    #[cfg(windows)]
+    use {crate::starlark::file_content::FileContentValue, tugger_common::testutil::*};
 
     #[test]
     fn test_new() -> Result<()> {
         let mut env = StarlarkEnvironment::new()?;
 
-        let v = env.eval("WiXMSIBuilder('prefix', 'name', '0.1', 'manufacturer')")?;
-        assert_eq!(v.get_type(), "WiXMSIBuilder");
+        let builder_value = env.eval("WiXMSIBuilder('prefix', 'name', '0.1', 'manufacturer')")?;
+        assert_eq!(builder_value.get_type(), "WiXMSIBuilder");
+        let builder = builder_value.downcast_ref::<WiXMsiBuilderValue>().unwrap();
+        assert_eq!(builder.inner.lock().unwrap().arch, "x64");
 
         Ok(())
     }
@@ -335,4 +496,37 @@ mod tests {
 
         Ok(())
     }
+
+    #[cfg(windows)]
+    #[test]
+    fn to_file_content() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        env.eval("msi = WiXMSIBuilder('prefix', 'name', '0.1', 'manufacturer')")?;
+        let value = env.eval("msi.to_file_content()")?;
+
+        assert_eq!(value.get_type(), FileContentValue::TYPE);
+
+        Ok(())
+    }
+
+    #[cfg(windows)]
+    #[test]
+    fn write_to_directory() -> Result<()> {
+        let mut env = StarlarkEnvironment::new()?;
+
+        let dest_dir = DEFAULT_TEMP_DIR
+            .path()
+            .join("wix-installer-write-to-directory");
+        let dest_dir_s = dest_dir.to_string_lossy().replace('\\', "/");
+
+        env.eval("msi = WiXMSIBuilder('prefix', 'name', '0.1', 'manufacturer')")?;
+        let value = env.eval(&format!("msi.write_to_directory('{}')", dest_dir_s))?;
+
+        assert_eq!(value.get_type(), "string");
+        let path = PathBuf::from(value.to_string());
+        assert!(path.exists());
+
+        Ok(())
+    }
 }
